/*
xferix -- Display dates in various formats

    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later or gcc 2.95.02 or later.  
    cl whenis.c -or- gcc whenis.c

	Usage is: whenis date fmt where date is a date and fmt gives the format
	slash stands for either a /, . or :
	fmt = ymd for yyyymmdd or yyyy/mm/dd
	fmt = mdy for mmddyyyy or mm/dd/yyyy
	fmt = dmy for ddmmyyyy or dd/mm/yyyy
	fmt = sec for seconds from 1970/01/01 
	fmt = jul for yyyyddd (year followed by day of year)
	format is required.

	This source is compatible with both Windows and Unix
*/


#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <time.h>
#include <stdlib.h>
#define BUFSIZE 64000
#include <stddef.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <dirent.h>

char *SearchDir(char *dirname, int size, char *filename, char quiet) {		
	DIR *dp;
	struct dirent *ep;
	FILE *ip;
	struct stat buf;
	int rc;
	char pathname[1024], *p;
	char pline[64];

	pathname[1023]=0;
	for (p=pathname;*p;*p++ = ' ');
	strcpy(pathname, dirname);
	if (strlen(dirname) > 63) {
		// take first 30 and last 30
		strncpy(pline,dirname,30);
		pline[30]=0;
		strcat(pline,"...");
		strcat(pline,dirname+strlen(dirname)-30);
	}

  	dp = opendir (dirname);
        if (dp != NULL) {
		// FPrintf(stderr,"%s\n",dirname);
		if (!quiet) {
			printf("%s\r",pline);
			fflush(stdout);
		}
		while (ep = readdir (dp)) {
			//puts(ep->d_name);
			strcpy(pathname,dirname);
			if (pathname[strlen(pathname)-1] != '/')
				strcat(pathname,"/");
			strcat(pathname,ep->d_name);
//			puts(pathname);
			rc = lstat(pathname,  &buf);
			if (rc==0) {
			// look for a file.  If found and it is the right size, open it.  If it can't be opened, look for another file
				if (buf.st_size >= size) {
					strcpy(filename,pathname);
					ip = fopen(filename,"r");
					if (ip != NULL) {
						fclose(ip);
						return filename;
					}
				}
				if (buf.st_mode & S_IFDIR) {
//					puts(" ");
			        	if (strcmp("..",ep->d_name) && strcmp(".",ep->d_name))
			        		if (buf.st_mode & S_IFLNK)
			        			;
			        		else 
							if (SearchDir(pathname, size, filename,quiet))
								return filename;
					else
						;
				}
			}
			else
                              ;
		}
		(void) closedir (dp);
	}
	else {
		// fprintf(stderr,"%s -- ",dirname);
	      	// perror ("Couldn't open the directory");
	}
	return NULL;
}


double startnow, endnow, confidence;
long minsize=0;

char *comma_delim(long double lno,char *buffer,char mb,char kb) {
/* returns the long integer in comma delimited format */
/* Assume buffer is long enough to hold the largest value */
/* mb = 1; return result in MB; kb = 1 return result in KB */
	char b1[36],*b; /* should be long enough to hold largest ltoa value */
	int i,n,l,k,c;
	long double no;
	no = lno;
	if (mb) {
		no = (lno / (1024*1024));
		if (no > -1.0 && no < 1.0) {
			kb = 1;
			mb = 0;
		}
	}

	if (kb) {
		no = (lno / 1024);
		if (no > -1.0 && no < 1.0) {
			no = lno;
			kb = 0;
		}
	}

	sprintf(b1,"%16.0f",no);
	for (b=b1;*b==' ';++b) ;  // Remove leading blanks

	l = strlen(b); /* l is length of string and index to 0 at end of string */
	n = l/3;	   /* n is number of commas to be inserted */
	if ((l % 3) == 0)
		--n;
	k = l + n;	   /* k is length of new string */
	buffer[k--] = b[l--];
	for (c=0,i=l; i >= 0; --i,--k) {
		buffer[k] = b[i];
		if (++c == 3 && i > 0) {
			buffer[--k] = ',';
			c = 0;
		}
	}

	if (mb)
		strcat(buffer,"MB");
	else if (kb)
		strcat(buffer,"KB");

	return buffer;
}

char *tgm(double val,char *valstr) {
	// reformat number so that it looks like xxx.xT. xxx.xM. xxx.xK. xxx.x
	double oneT =   1000000000000.;
	double oneG =   1000000000.;
	double oneM =   1000000;
	double oneK =   1000;
	if (val >= oneT) 
		sprintf(valstr,"%.1lfT",val/oneT);
	else if (val >= oneG)
		sprintf(valstr,"%.1lfG",val/oneG);
	else if (val >= oneM)
		sprintf(valstr,"%.1lfM",val/oneM);
	else if (val>= oneK)
		sprintf(valstr,"%.1lfK",val/oneK);
	else
		sprintf(valstr,"%.1f",val);
	return valstr;
}

// XFERIX -- Xferrate for Unix systems (or compatibles)
// Should also work under DOS and Windows
// because it uses standard C library only
//
void main(int argc,char *argv[]) {
	double size,f_elapsed,bps;
	time_t start,stop,prev;
	unsigned long bytes;
	unsigned si,bufsi,dummy;
        char inputfile[1024],*buffer,thyme[4],more,temp_bps[64],temp_size[64], debugbuf[256];
	struct stat statbuf;
        struct timeval tv_start, tv_end;
	struct timezone tz;
	int in,rci,rc;
	long int elapsed, opentime;
	int arg,quiet=0;
	int error_count = 0, max_errors = 0, short_blocks = 0;
	FILE *hlp;
	
	if (argc < 2 ||  strcmp(argv[1],"--help") ==  0 || strcmp(argv[1],"-?") == 0) {
		buffer = getenv("HOME");
		if (buffer != NULL) {
			strcpy(inputfile,"less ");
			strcat(inputfile, buffer);
			if (inputfile[strlen(inputfile)-1] != '/')
				strcat(inputfile,"/");
			strcat(inputfile,"temp");
			mkdir(inputfile+5,S_IRWXU|S_IRWXG);
			strcat(inputfile,"/xferix.hlp");
			fprintf(stderr,"%s\n",inputfile);
			hlp = fopen(inputfile+5,"w");
		}
		if (hlp == NULL)
			hlp = stderr;
			
		fprintf(hlp,"xferix Version 1.30, %s\n",__DATE__);
  	    	fprintf(hlp,"\nCopyright 2003, Norm London, London Computing.\nnorm_london@londoncomputing.com\n\n");
        	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
		fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
		fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
		fprintf(hlp,"(at your option) any later version.\n\n");
		fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
		fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
		fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
		fprintf(hlp,"GNU General Public License for more details.\n\n");
		fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
		fprintf(hlp,"along with this program; if not, write to the Free Software\n");
		fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
		//fprintf(hlp,"Press the Enter key to continue");
		// getchar();

		fprintf(hlp,"xferix measures the transfer rate of a hard drive or lan connection.\n");
		fprintf(hlp,"Usage is:\n");
		fprintf(hlp,"xferix --help -or- xferix -?       prints this help message.\n");
		fprintf(hlp,"xferix [-q]  [-snnn] filename [bufsize [errorkt]]     measures transfer rates.\n");
		fprintf(hlp,"Where: filename is the name of the file to be read.\n");
		fprintf(hlp,"       filename should be at least 32KB in size.\n");
		fprintf(hlp,"If the filename points to a directory and the -snnn parameter is specified, \n");
		fprintf(hlp,"xferix will use the first file it finds in the specified path that is at least\n");
		fprintf(hlp,"nnn Bytes in size and can be read by xferix. \n\n");

		fprintf(hlp,"bufsize is the size of the read buffer in Bytes.  64 <= bufsize <= 65535\n");
		fprintf(hlp,"-OR- the size of the read buffer in KiloBytes.     1 <= bufsize <  63\n");
		fprintf(hlp,"if bufsize omitted, 64000 is used.\n");
		fprintf(hlp,"errorkt is the number of read errors to ignore before quitting, default 0\n");
		fprintf(hlp,"The output of XFERIX shows a continuous display of the number.\n");
		fprintf(hlp,"Size (bytes read), Time (in seconds), Xfer Rate (Transfer Rate).\n");
		fprintf(hlp,"For most accurate results, files should be large enought that\n");
		fprintf(hlp,"transfer times exceep 100 seconds.\n");
		fprintf(hlp,"Xferix may also report on:\n");
		fprintf(hlp,"Short Blks  (Number of reads returning less than the block size.)\n");
		fprintf(hlp,"             One short block is normal when the file size is not\n");
		fprintf(hlp,"             a multiple of the block size and will not be reported.\n");
		fprintf(hlp,"             Short blocks are most prevalent on slow LAN links where\n");
		fprintf(hlp,"             a timeout occurs before the entire block is read.\n");
		fprintf(hlp,"Error Count (Number of reads returning errors.\n");
		fprintf(hlp,"-q -- Quiet Mode.  Do display statistics every second.\n");
		
		if (hlp != stderr) {
			fclose(hlp);
			fprintf(stderr,"%s\n",inputfile);
			system(inputfile);
		}
		exit(1);
	}
	arg = 0;
        if (argv[arg+1][0] == '-')
                if (toupper(argv[arg+1][1]) == 'Q') {
                        quiet = 1;
                        arg = 1;
                }
                else if (toupper(argv[arg+1][1]) == 'S') {
                        minsize = atol(argv[arg+1]+2);
                        arg = 1;
                }
        if (argv[arg+1][0] == '-')
                if (toupper(argv[arg+1][1]) == 'Q') {
                        quiet = 1;
                        arg = 1;
                }
                else if (toupper(argv[arg+1][1]) == 'S') {
                        minsize = atol(argv[arg+1]+2);
                        arg = 1;
                }
	if (argc > (2+arg))
		bufsi = atoi(argv[arg+2]);
	else
		bufsi = 0;

	if (argc > (arg+3))
		max_errors = atoi(argv[arg+3]);

	if (bufsi == 0 || bufsi > 65535)
		bufsi = BUFSIZE;
	else if (bufsi >= 1 && bufsi <= 63)
		bufsi *= 1024;

	printf("Buffer Size = %u Bytes, Allow %d Errors.\n",bufsi,max_errors);
	if ((buffer = (char *)malloc(bufsi)) == NULL) {
		fprintf(stderr,"Insufficient Memory\n");
		exit(1);
	}
	
	strcpy(inputfile, argv[arg+1]);
	rc = stat(inputfile,&statbuf);
    	if (rc == 0 && (statbuf.st_mode & S_IFDIR) && minsize>0) {
                if (SearchDir(argv[arg+1], minsize, inputfile, quiet)) {
			printf("Read file: %s\n",inputfile);
		}
		else {
			fprintf(stderr,"Unable to find a file in the path: %s that is at least %ld in size.\n",
			argv[arg+1],minsize);
			exit(4);
		}
	}
	// know file name, open it
	rc = gettimeofday(&tv_end,&tz);
	in = open(inputfile,O_RDONLY);
	if (in == -1) {
		perror(inputfile);
		exit(1);
	}


	// startnow = (double)clock() / CLOCKS_PER_SEC;
	start = time(&start);
	rc  = gettimeofday(&tv_start,&tz);
	opentime = (tv_start.tv_sec  - tv_end.tv_sec) * 1000000 + (tv_start.tv_usec - tv_end.tv_usec);
	prev = start;
	// printf("START = %ld, STARTNOW = %.2f\n", start,startnow);
	size = 0;
	si = read(in,buffer,bufsi);
	bytes = si;
	while (si && error_count <= max_errors) {
		size += (si == 0xffff ? 0 : si);
		si = read(in,buffer,bufsi);
		stop = time(&stop);
		if (si) {
			if (si == 0xFFFF && si != bufsi) {
				++error_count;
			}
			else if (si < bufsi)
			    ++short_blocks;
			if (stop > start && stop > prev && quiet == 0) {
				fprintf(stderr,"\r                                                                        ");
				fprintf(stderr," \rSize=%.0f, Time = %ld, Xfer Rate = %.0f",
					size,stop-start,size/(stop-start));
				if (short_blocks)
					fprintf(stderr,", Short Blk = %d",short_blocks);
				if (error_count)
					fprintf(stderr,", Errors = %d",error_count);
				prev = stop;
			}
		}
	}
	size += (si == 0xFFFF ? 0 : si);
	// endnow = (double)clock()/CLOCKS_PER_SEC;
	stop = time(&stop);
	rc = gettimeofday(&tv_end,&tz);
	
	startnow = start;
	endnow = stop;
        elapsed = (tv_end.tv_sec - tv_start.tv_sec)*1000000 + tv_end.tv_usec - tv_start.tv_usec;
	f_elapsed =  (float)elapsed/1000000;
	bps = size / f_elapsed;
	// printf("\nelapsed = %ld",elapsed);
	// printf("\nSTOP=%ld, ENDNOW=%.2f\n",stop,endnow);
	if (stop > start && elapsed > 0) {
		confidence = 500.0 / (elapsed);
		endnow += .5;
		printf("\nOpen(us)=%ld, Size=%sB, Time(Sec)=%.2f, Rate = %sBPS\n",
		opentime,tgm(size,temp_size),f_elapsed,tgm(bps,temp_bps));
//		printf("\nSize(Bytes)=%.0f, Time(Seconds)=%.2f, Transfer Rate(BPS) = %.0f.\n",
//			size,f_elapsed,size/f_elapsed);
		if (short_blocks > 1) 
			printf("%d Short Blocks.  ",short_blocks);
		if (error_count)
			printf("%d Errors",error_count);
		if (short_blocks > 1 || error_count > 0)
			putchar('\n');
	}
	else
		printf("\nFile too small to time.  Please select a larger file.\n");
	close(in);
	exit(0);
}
